// Builtin.cc - ColdStore interface to C++ functions
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms

static char *id __attribute__((unused))="$Id: Builtin.cc,v 1.3 1999/03/13 00:40:28 colin Exp $";

#include <iostream.h>
#include <elf.hh>

#include "Data.hh"
#include "Store.hh"
#include "Task.hh"

#include "String.hh"
#include "Tuple.hh"
#include "Error.hh"

#include "Builtin.hh"

// create a Builtin from a Csym
Builtin::Builtin(Csym *_sym)
    : String(_sym->symname()), elf(_sym->in), sym(_sym)
{}

// construct from constructor args
// expects a 2Tuple or List of form (builtin name, file-name)
Builtin::Builtin(const Slot &arg)
    : elf(Elf::findElf((int)(arg->length()) == 2
                       ?(char*)(arg[1])
                       : COLDSTORE)),
      sym(elf->findSym(arg[0]))
{
    if (!currTask()->isSys()) {
        throw new Error("illegal", arg, "Constructing Builtin is privileged");
    }
}

// constructor args to recreate object
Slot Builtin::toconstruct() const
{
    if (!strcmp(fname(), COLDSTORE)) {
        return new String(name());
    } else {
        return new (2) Tuple(2, new String(sym->symname()),
                             new String(elf->name()));
    }
}

Builtin::~Builtin()
{}

//////////////////////////////////////////////////
// Builtin static construction

//////////////////////////////////////////////////
// Builtin accessors

const char *Builtin::name() const {
    return sym->symname();
}

const char *Builtin::fname() const {
    return elf->name();
}

bool Builtin::isLoaded() const {
    return sym->isLoaded();
}

//////////////////////////////////////////////////
// Builtin feature extractors

///////////////////////////////
// Coldmud Interface

// structural

// Identity functions
Data *Builtin::clone(void *store) const
{
    return (Data*)this;
}

Data *Builtin::mutate(void *where) const
{
    return (Data*)this;
}

// Is the Builtin loaded?
bool Builtin::truth() const
{
    // handle will be non-0 if Builtin's been loaded.
    return sym->isLoaded();
}

// Order by the string order of the file and builtin names
int Builtin::order(const Slot &arg) const
{
    int cmp = typeOrder(arg);
    if (cmp)
        return cmp;
    
    Builtin *other = (Builtin*)arg;

    cmp = strcmp(fname(), other->fname());
    if (!cmp)
        cmp = strcmp(name(), other->name());
    return cmp;
}

// sequence

// get length as sequence
int Builtin::length() const
{
    return 5;
}

// get C++ symbol as Builtin
Slot Builtin::slice(const Slot &from, const Slot &to)
{
    if (to)
        throw new Error("value", to, "Builtin can't handle slices");
    
    // give access to builtin names
    switch((int)from) {
    case 0:
        return name();
    case 1:
        return fname();
    case 2:
        return isLoaded()?1:0;
    case 3:
        return sym->len();
    case 4: {
        switch (sym->type()) {
        case Csym::FUNCTION:
            return "function";
        case Csym::DATA:
            return "data";
        case Csym::VTBL:
            return "vtbl";
        case Csym::UNKNOWN:
            return "unknown";
        default:
            return "impossible";
        }
    }
    default:
        throw new Error("range", from, "Builtin has no such entry");
    }
}

const void *Builtin::vtbl() const {
    try {
        return sym->vtype();
    } catch (char *err) {
        throw new Error("builtin", this, err);
    }
}

/////////////////////////////////////
// functional

// call Builtin
Slot Builtin::call(const Slot &args) const
{
    if (!sym)
        throw new Error("builtin", this, "no associated Csym");
    if (sym->type() != Csym::FUNCTION)
        throw new Error("builtin", this, "Csym not FUNCTION type");

    try {
        return (Data*)(sym->call((void*)(Data *)args));
    } catch (char *err) {
        throw new Error("builtin", this, err);
    }
}

ostream &Builtin::dump(ostream& out) const
{
    char *ty;
    out << typeId() << ':';
    String::dump(out);

    switch (sym->type()) {
    case Csym::FUNCTION:
        ty = "FUNCTION"; break;
    case Csym::DATA:
        ty = "DATA"; break;
    case Csym::VTBL:
        ty = "VTBL"; break;
    case Csym::UNKNOWN:
        ty = "UNKNOWN"; break;
    default:
        ty = "ERRONEOUS"; break;
    }

    out << " = " << sym->symname()
        << " type: " << ty;

    return out;
}

