// Dict - ordered mapping of key to value
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms

static char *id __attribute__((unused))="$Id: Dict.cc,v 1.3 1999/03/13 00:40:27 colin Exp $";

#include "Data.hh"
#include "Store.hh"

#include "String.hh"
#include "Tuple.hh"
#include "Error.hh"
#include "List.hh"

#include "Dict.hh"

class SlotPair
    : public Memory
{
    
protected:
    Slot _key;
    mutable Slot _value;

public:
    bool operator == (SlotPair &p);
    bool operator != (SlotPair &p);
    bool operator < (SlotPair &p);
    bool operator <= (SlotPair &p);
    bool operator >= (SlotPair &p);
    bool operator > (SlotPair &p);
    int order(const SlotPair *arg) const;
    
    // Constructors
    
    // used to initialize to NULL - UGLY
    SlotPair()
        : _key((Data*)0), _value((Data*)0) {}

    // a valueless pair - used to search
    SlotPair(const Slot &k)
        : _key(k), _value((Data*)0) {}

    SlotPair(const SlotPair &p)	// Copy Constructor
        : _key(p._key), _value(p._value) {}

    SlotPair(const Slot &k, const Slot &v)
        : _key(k), _value(v) {}

    const SlotPair &operator =(const SlotPair &p);
    const int operator =(const int i);

    // accessors
    Slot &key() {
        return _key;
    }

    Slot &value() {
        return _value;
    }
};

bool SlotPair::operator == (SlotPair &p) {
    return (_key == p._key);
}

bool SlotPair::operator != (SlotPair &p) {
    return (_key != p._key);
}

bool SlotPair::operator < (SlotPair &p) {
    return (_key < p._key);
}

bool SlotPair::operator <= (SlotPair &p) {
    return (_key <= p._key);
}

bool SlotPair::operator >= (SlotPair &p) {
    return (_key >= p._key);
}

bool SlotPair::operator > (SlotPair &p) {
    return (_key > p._key);
}

const SlotPair &SlotPair::operator =(const SlotPair &p)
{
    _key = p._key;
    _value = p._value;
    return p;
}

const int SlotPair::operator =(const int i)
{
    _key = (Data*)0;
    _value = (Data*)0;
    return i;
}

int SlotPair::order(const SlotPair *arg) const
{
    return _key->order(arg->_key);
}

template class BTree<SlotPair>;
template class PathStruct<SlotPair>;
template class BTreeNode<SlotPair>;
template class BTreeIterator<SlotPair>;

Dict::Dict(const Slot &init)
    : tree() //BTree<SlotPair>()
{
    int l = init->length();
    for (int i = 0; i < l; i+=2) {
        Slot key = init[i];
        Slot val = init[i+1];
        tree.BTree<SlotPair>::Add(SlotPair(key, val));
    }
}

Dict::Dict(int order)
    : tree(order) //BTree<SlotPair>(order)
{
}

Dict::Dict(const Dict &o, int order)
    : tree(order) // BTree<SlotPair>(order)
{
    //BTreeIterator<SlotPair> it(o);
    DictIt it(o);
    while (it.More()) {
        const Slot &key = (it.Current());
        const Slot &val = (it.Next());
        Add(key, val);
    }
}

Dict::Dict(const Slot &pairs, int order)
    : tree(order) //BTree<SlotPair>(order)
{
    Slot p = pairs->toSequence();
    for (int i = 0; i < p->length(); i+=2) {
        tree.BTree<SlotPair>::Add(SlotPair(p[i], p[i+1]));
    }
}

Dict::~Dict()
{}

// structural

// Uses the copy constructor
Data *Dict::clone(void *where) const
{
    return new Dict(*this);
}

// returns a mutable copy of this
Data *Dict::mutate(void *where) const
{
    if (refcount() > 1)
        return clone(where);
    else
        return (Data*)this;
}

/////////////////////////////////////
// object

// the object's truth value
bool Dict::truth() const
{
    return tree.Size() != 0;
}

// constructor args to recreate object
Slot Dict::toconstruct() const
{
    int size = 2 * tree.Size();
    if (size) {
        Slot t = new (size) Tuple(-size);
        //Slot::init(t->content(), size);
        
        BTreeIterator<SlotPair> it(tree);
        for (int i = 0; it.More(); i+=2) {
            SlotPair *item = it.Next();
            (*(Tuple*)t)[i] = item->key();
            (*(Tuple*)t)[i+1] = item->value();
        }
        return t;
    } else {
        return new List(0);
    }
}

// 1,0,-1 depending on order
int Dict::order(const Slot &arg) const
{
    BTreeIterator<SlotPair> it1(tree);
    BTreeIterator<SlotPair> it2(((Dict*)arg)->tree);
    while (it1.More() && it2.More()) {
        SlotPair *item1 = it1.Next();
        SlotPair *item2 = it2.Next();
        int result = item1->order(item2);
        if (result)
            return result;
    }
    return 0;
}

// 1,0,-1 depending on order
bool Dict::equal(const Slot &arg) const
{
    if ((void*)&arg == (void*)this)
        return true;
    
    BTreeIterator<SlotPair> it1(tree);
    BTreeIterator<SlotPair> it2(((Dict*)arg)->tree);
    while (it1.More() && it2.More()) {
        SlotPair *item1 = it1.Next();
        SlotPair *item2 = it2.Next();
        if (*item1 != *item2)
            return false;
    }
  return true;
}


/////////////////////////////////////
// sequence

// length as sequence
int Dict::length() const
{
    return tree.Size();
}

// concatenate two sequences - Merge two Dicts
Slot Dict::concat(const Slot &arg)
{
    if (arg->typeEQ(store->typeDict)) {
        // we're concatenating two Dicts
        Dict *m = (Dict*)mutate();
        Dict *other = arg;
        
        BTreeIterator<SlotPair> it(other->tree);
        while (it.More()) {
            SlotPair *item = it.Next();
            tree.BTree<SlotPair>::Add(SlotPair(item->key(), item->value()));
        }
        return m;
    } else {
        // construct an order, then merge
        Slot tmp = new Dict(arg);
        return concat(tmp);
    }
}

// dyadic `[]', range
Slot Dict::slice(const Slot &from, const Slot &len) const
{
    // a range of items
    Slot l = new List;
    
    BTreeIterator<SlotPair> it(tree);
    it.BeginFrom(SlotPair(from));
    while (it.More()) {
        SlotPair *item = it.Next();
        
        // break at end of subrange
        if (item->key() > len)
            break;
        
        // add the next item to the list
        l->add(item->key());
        l->add(item->value());
    }
    return l;
}

// dyadic `[]', item
Slot Dict::slice(const Slot &from) const
{
    // a single item
    SlotPair probe(from);
    SlotPair *found = tree.Find(probe);
    if (found == (SlotPair*)0)
        throw new Error("dictnf", from, "key not found");
    else {
        Slot l = new List;
        l->add(found->key());
        l->add(found->value());
        return l;
    }
}

// dyadic `search', matching subrange
Slot Dict::search(const Slot &what) const
{
    SlotPair probe(what);
    SlotPair *found = tree.Find(probe);
    if (found == (SlotPair*)0)
        return (Data*)0;
    else {
        return found->value();
    }
}

// delete item from Dict
Slot Dict::del(const Slot &from)
{
    Dict *m = (Dict*)mutate();
    if (m->tree.Find(from)) {
        m->tree.Remove(from);
    } else {
        throw new Error("dictnf", from, "key not found");
    }
    return m;
}

// delete subrange
Slot Dict::del(const Slot &from, const Slot &len)
{
    // Delete item(s) from Dict
    Dict *m = (Dict*)mutate();
    BTreeIterator<SlotPair> it(m->tree);
    it.BeginFrom(SlotPair(from));
    while (it.More()) {
        SlotPair *item = it.Next();
        
        // break at end of subrange
        if (item->key() > len)
            break;
        
        m = m->del(item->key());
    }
    
    return m;
}

// replace subrange
Slot Dict::replace(const Slot &from, const Slot &to, const Slot &value)
{
    Dict *m = (Dict*)mutate();

    // Assign a range of items
    BTreeIterator<SlotPair> it(tree);
    Slot values = value->iterator();
    it.BeginFrom(SlotPair(from));
    while (it.More()) {
        SlotPair *item = it.Next();
        
        // break at end of subrange
        if (item->key() > to)
            break;
        
        m = m->replace(item->key(), values->Current());		// replace the current value
        // move to the next assignment value
        if (values->More()) {
            values->Next();
        } else {
            if (it.More()) {
                throw new Error("range", item->key(), "Values exhausted");
            }
        }
    }
    return m;
}

// replace Item
Slot Dict::replace(const Slot &key, const Slot &value)
{
    Dict *m = (Dict*)mutate();
    
    // Assign a single item
    SlotPair *found = m->tree.Find(SlotPair(key));
    if (!found){
        // unknown symbol - create and add it
        throw new Error("dictnf", key, "key not found");
    } else {
        // bind new value
        found->value() = value;
    }
    return m;
}

// insert Item
Slot Dict::insert(const Slot &key, const Slot &value)
{
    Dict *m = (Dict*)mutate();
    
    // Assign a single item
    SlotPair *found = m->tree.Find(SlotPair(key));
    if (!found){
        // unknown symbol - create and add it
        m->tree.Add(SlotPair(key, value));
    } else {
        // bind new value
        found->value() = value;
    }
    return m;
}

// insert Item
Slot Dict::insert(const Slot &value)
{
    Dict *m = (Dict*)mutate();
    
    // Assign a single item
    SlotPair *found = m->tree.Find(SlotPair(value));
    if (!found){
        // unknown symbol - create and add it
        m->tree.Add(SlotPair(value));
    } else {
        // bind new value
        found->value() = value;
    }
    return m;
}

Slot Dict::replace(const Slot &value)
{
    return Data::replace(value);	// in fact, it throws
}

// return an iterator
Slot Dict::iterator() const
{
    return new DictIt(this);
}

/////////////////////////////////////
// map

// return map as sequence
Slot Dict::toSequence() const
{
    // a range of items
    Slot l = new List;
    
    BTreeIterator<SlotPair> it(tree);
    while (it.More()) {
        SlotPair *item = it.Next();
        
        // add the next item to the list
        l->check();
        l = l->add(item->key());
        l = l->add(item->value());
        // Note: we can get away with not assigning to l
        // because its refcount is 1
    }
    return l;
}

// Local
bool Dict::Add(const Slot &key, const Slot &value)
{
    SlotPair p(key, value);
    return tree.Add(p);
}

SlotPair *Dict::Find(const Slot &key) const
{
    SlotPair p(key);
    return tree.Find(p);
}

bool Dict::FindV(const Slot &key, Slot &value) const
{
    SlotPair p(key, value);
    SlotPair *found = tree.Find(p);
    if (found) {
        value = found->value();
    }
    return found != NULL;
}

void Dict::Replace(SlotPair *entry, const Slot &value)
{
    entry->value() = value;
}


//////////////////////////////////
// DictIt - Iterator over Dicts

DictIt::DictIt(const Dict *o)
    : BTreeIterator<SlotPair>(o->tree),
      _order(o),
      min((Data*)0),
      max((Data*)0) {}

DictIt::DictIt(const Dict &o)
    : BTreeIterator<SlotPair>(o.tree),
      _order(&o),
      min((Data*)0),
      max((Data*)0) {}

DictIt::DictIt(const DictIt &oit)
    : BTreeIterator<SlotPair>(((Dict*)(oit._order))->tree),
      _order(oit._order),
      min((Data*)0),
      max((Data*)0) {}

DictIt::~DictIt() {}

// structural

// Uses the copy constructor
Data *DictIt::clone(void *where) const
{
    return new DictIt(*this);
}

// returns a mutable copy of this
Data *DictIt::mutate(void *where) const
{
    return clone(where);
}

// object

// the object's truth value
bool DictIt::truth() const
{
    return More();
}

// constructor args to recreate object
Slot DictIt::toconstruct() const
{
    return _order;
}

// construct from constructor args
Slot DictIt::construct(Slot &arg)
{
    if (!arg->typeEQ(store->typeDict))
        throw new Error("type", arg, "DictIt may only be constructed from an Dict");
    return new DictIt((Dict*)arg);
}

// 1,0,-1 depending on order
int DictIt::order(const Slot &arg) const
{
    if (typeOrder(arg))
        return typeOrder(arg);
    if ((((DictIt*)arg)->_order != _order))
        return _order->order(((DictIt*)arg)->_order);
    else 
        return CurrentRank() - ((DictIt*)arg)->CurrentRank();
}

// depending on equality
bool DictIt::equal(const Slot &arg) const
{
    return arg->typeEQ(store->typeDictIt)
        && (((DictIt*)arg)->_order == _order)
        && (((DictIt*)arg)->CurrentRank() == CurrentRank());
}

// iterator

// predicate - is object an iterator
bool DictIt::isIterator()
{
    return true;
}

Slot DictIt::Current()
{
    if (CurrentRank() == -1) {
        BTreeIterator<SlotPair>::Reset();	// move to start
    }
    SlotPair *curr = ((Dict*)_order)->tree.ItemWithRank(CurrentRank());
    if (!curr)
        throw new Error("exhausted", this, "Iterator exhaustion");
    return new (2) Tuple(2, &(curr->key()), &(curr->value()));
}

bool DictIt::More() const
{
    return BTreeIterator<SlotPair>::More();
}

Slot DictIt::Next()
{
    if (BTreeIterator<SlotPair>::More())
        BTreeIterator<SlotPair>::Next();	// move to start  
    else
        throw new Error("break", this, "Iterator exhaustion");
    return Current();
}

