// Integer - ColdStore interface to 32 bit integer type
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms

static char *id __attribute__((unused))="$Id: Integer.cc,v 1.2 1999/03/11 06:51:03 colin Exp $";
//#define DEBUGLOG 1
#include <stdlib.h>
#include <stdio.h>

#include <qvmm.h>
#include <Q.hh>
#include "Data.hh"
#include "Store.hh"

#include "Vector.hh"
#include "String.hh"
#include "Tuple.hh"
#include "List.hh"
#include "Error.hh"
#include "Integer.hh"

// instantiate templates
template class Q<iPool>;
template class Qh<iPool>;

struct PoolCell {
    union {
        PoolCell *next;	// next free PoolCell
        iPool *home;	// the IPool this PoolCell resides in
    };
    Integer *content;
    
    PoolCell()
        : next(this + 1),
          content(new Integer(0, this))
        {}
    ~PoolCell() {
        delete content;
    }
};

class iPool
    : public Memory,
      public Q<iPool>
{
    static Qh<iPool> all;	// Q of all Pools of Integer
    static const size_t def_size = (Mmap::page_size / sizeof(Integer*))-1;
    PoolCell content[def_size];
    PoolCell *freeH;

    Integer *get();

public:
    iPool();
    ~iPool();

    static Integer* Get();
    void release(Integer *el);
};

iPool::iPool()
    : Q<iPool>(all),
      freeH(content)
{
    DEBLOG(cerr << "Pool new " << this
           << " size " << def_size
           << '\n');
    content[def_size - 1].next = NULL;
}

iPool::~iPool()
{
    throw "Trying to delete Pool";
}

Integer *iPool::get()
{
    DEBLOG(cerr << "iPool " << this
           << '(' << next() << ',' << prev() << ") get ");
    if (freeH) {
        // we have a free item
        PoolCell *gotit = freeH;
        freeH = freeH->next;
        DEBLOG(cerr << gotit->content << '\n');

        gotit->home = this;
        return gotit->content;
    }
    DEBLOG(cerr << "NULL\n");
    return NULL;
}

Integer* iPool::Get()
{
    if (!all) {
        iPool *p =  new iPool();
        return p->get();
    } else {
        // find a Pool with free
        for (Qi<iPool> pi(all); pi; pi++) {
            Integer *found = ((iPool*)pi)->get();
            if (found)
                return found;
        }

        // no Pool has free, create new Pool
        DEBLOG(cerr << " new Pool\n");
        iPool *p = new iPool();
        return p->get();
    }
}

void iPool::release(Integer *el)
{
    DEBLOG(cerr << "Pool " << this << " release " << el << '\n');
    
    // add element to the free linked list
    el->_home->next = freeH;
    freeH = el->_home;
}

Integer::Integer(const Slot &init)
    : _value(((Integer*)init)->value()),
      _home(NULL)
{
    DEBLOG(cerr << "Integer " << this
           << " value " << _value << '\n');
}

Integer::Integer(long value, PoolCell *home)
    : _value(value),
      _home(home)
{
    DEBLOG(cerr << "Integer " << this
           << " in " << _home
           << " value " << _value << '\n');
}

Integer::~Integer()
{
    DEBLOG(cerr << "Destroy Integer " << this
           << " in " << _home
           << '\n');
    if (_home) {
        _home->home->release(this);	// put back on the freelist
    }
}

Integer *Integer::getInteger(long value)
{
    Integer *result = iPool::Get();
    if (!result) {
        throw "Can't get iPool Integer";
    }
    result->_value = value;
    result->upcount();
    return result;
}

long Integer::operator = (long value)
{
    return _value = value;
}

long Integer::value()
{
    return _value;
}

Data *Integer::clone(void *where) const
{
    DEBLOG(cerr << "clone Integer " << this
           << " in " << _home
           << " value " << _value << '\n');
    if (_home) {
        return getInteger(_value);
    } else {
        return new (where) Integer(_value);
    }
}

void Integer::check(int) const {}

void Integer::operator delete(void* ptr)
{
    if (static_cast<Integer*>(ptr)->_home) {
        // This Pool allocated: do nothing at all
    } else {
        // this is a freely allocated Integer
        Memory::operator delete(ptr);
    }
}

// 1,0,-1 depending on order
int Integer::order(const Slot &arg) const
{
    int cmp = typeOrder(arg);
    if (cmp) {
        return cmp;
    }

    int argi = (int)arg;
    return _value - argi;
}

// the object's truth value
bool Integer::truth() const
{
    return _value != 0;
}

// constructor args to recreate object
Slot Integer::toconstruct() const
{
    char buffer[32];
    sprintf(buffer, "%ld", _value);
    return new String(buffer);
}

ostream &Integer::dump(ostream& out) const
{
    out << _value;
    return out;
}

// equality predicate
bool Integer::equal(const Slot &arg) const
{
    //cerr << _value << " == " << (int)arg << '\n';
    return _value == (int)arg;
}


// arithmetic
// monadic `+', absolute value
Slot Integer::positive()
{
    return abs(this);
}

// monadic `-', negative absolute value
Slot Integer::negative()
{
    return -_value;
}

// dyadic `+', add
Slot Integer::add(const Slot &arg)
{
    return _value + (int)arg;
}

// dyadic `-', subtract
Slot Integer::subtract(const Slot &arg)
{
    return _value - (int)arg;
}

// dyadic `*', multiply
Slot Integer::multiply(const Slot &arg)
{
    return _value * (int)arg;
}

// dyadic '/', divide
Slot Integer::divide(const Slot &arg)
{
    return _value / (int)arg;
}

// dyadic '%', modulo
Slot Integer::modulo(const Slot &arg)
{
    return _value & (int)arg;
}

// bitwise
// unary '~', invert
Slot Integer::invert()
{
    return ~_value;
}

// dyadic '&', bitwise and
Slot Integer::and(const Slot &arg)
{
    return _value & (int)arg;
}

// dyadic '^', bitwise xor
Slot Integer::xor(const Slot &arg)
{
    return _value ^ (int)arg;
}

// dyadic '|', bitwise or
Slot Integer::or(const Slot &arg)
{
    return _value | (int)arg;
}

// dyadic '<<', left shift
Slot Integer::lshift(const Slot &arg)
{
    return _value << (int)arg;
}

// dyadic '>>', right shift
Slot Integer::rshift(const Slot &arg)
{
    return _value >> (int)arg;
}


// misc Integer transforms to implement Builtins
Integer *Integer::random()
{
    Data::Unimpl("Stub - return a random integer");
    return (Integer*)0;
}

Integer *Integer::abs(Integer*)
{
    Data::Unimpl("Stub - return absolute value of integer");
    return (Integer*)0;
}
