// List.cc - Coldstore Lists
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms

static char *id __attribute__((unused))="$Id: List.cc,v 1.3 1999/03/13 00:40:28 colin Exp $";

#include "Data.hh"
#include "Store.hh"

#include "String.hh"
#include "Tuple.hh"
#include "Error.hh"
#include "List.hh"

template class Segment<Slot>;
template class Vector<Slot>;
template ostream& operator<< (ostream&, const Vector<Slot>&);

unsigned int Vector<Slot>::minimum_size = 4;
unsigned int Vector<Slot>::extra_allocation = 4;
unsigned int Vector<Slot>::minimum_occupancy = 4;


// Constructors
List::List(int size)
    : Vector<Slot>(size)
{
}

List::List(const Slot *contentT, int additional)
  : Vector<Slot>(contentT, additional)
{
}

List::List(const Slot &contentT, int repetition)
  : Vector<Slot>(contentT, repetition)
{
}

List::List(const List *contentT)
  : Vector<Slot>(contentT)
{
}

List::List(const List *contentT, int start, int l)
  : Vector<Slot>(contentT, start, l)
{
}

List::List(const List &contentT)
  : Vector<Slot>(&contentT)
{
}

List::List(const List &contentT, int start, int l)
  : Vector<Slot>(contentT, start, l)
{
}

List::List(const Tuple *contentT, int start, int l)
  : Vector<Slot>((TupleBase<Slot>*)contentT, start, l)
{
}

List::List(const Tuple *contentT)
  : Vector<Slot>((TupleBase<Slot>*)contentT)
{
}

List::~List()
{}

/* factor
 * returns a list of 3 lists: [elements only of list1, intersection, elements only of list2]
 */
Slot List::factor(const List *l) const {
  Slot dl1 = new List(this);
  Slot inter = new List((int)length() + (int)(l->length()));
  Slot dl2 = new List(l);
    
#ifndef EDEBUG
  Vector<Slot>::check();
  ((List*)l)->Vector<Slot>::check();
#endif

  for (int i = dl2->length(); i >= 0; i--) {
    Slot pos = dl1->search(dl2[i]);
    if (pos) {
      // in intersection - move it and remove it
      int p = pos;
      inter->add(dl2->slice(i));
      ((List*)dl1)->Vector<Slot>::del(p);
      ((List*)dl2)->Vector<Slot>::del(i);
    }
  }

  // format up the return value
  Slot factor = new List(3);
  factor->add(dl1);
  factor->add(inter);
  factor->add(dl2);

  return factor;
}
  
/////////////////////////////////////
// structural

// Uses the copy constructor
Data *List::clone(void *where) const
{
  return new (where) List(*this);
}

// returns a mutable copy of this
Data *List::mutate(void *where) const
{
  if (refcount() > 1)
    return clone(where);
  else
    return (Data*)this;
}


void List::check(int) const
{
  validGuard();
  Vector<Slot>::check();
}

/////////////////////////////////////
// object

// the object's truth value
bool List::truth() const
{
  return !(Vector<Slot>::empty());
}

// constructor args to recreate object
// an identity, for Lists
Slot List::toconstruct() const
{
  return mutate();
}

// construct from constructor args
Slot List::construct(const Slot &arg)
{
  return new List((List*)arg);
}

ostream &List::dump(ostream& out) const
{
    if (this != (List*)0) {
        Slot i(iterator());
        out << "{";
        while (i->More()) {
            i->Current().Dump(out);
            i = i->Next();
            if (i->More()) {
                out << ", ";
            }
        }
        out << "}";
    } else {
        out << "NULL";
    }
    return out;
}

// 1,0,-1 depending on order
int List::order(const Slot &arg) const
{
  int cmp = typeOrder(arg);
  if (cmp)
    return cmp;

  List *l = arg;
  if (this == l)
    return true;

#ifdef EDEBUG
  check();
  l->check();
#endif

  // Lists can only be equal if they're of the same length.
  int len = (int)(l->length()) <? (int)(length());

  // See if any elements differ.
  for (int i = 0; i < len; i++) {
    int diff = element(i)->order(l->element(i));
    if (diff != 0)
      return diff;
  }

  // they're equal to the end of one list, which is longer?
  return length() - (l->length());
}

//  Predicate: are lists equal?
bool List::equal(const Slot &arg) const
{
  // typed equality
  if (typeOrder(arg))
    return false;

  List *l = arg;
  if (this == l)
    return true;	// == is a reflexive relation

  // Lists can only be equal if they're of the same length.
  int len = Vector<Slot>::Length();
  if (l->Vector<Slot>::Length() != len)
    return false;

  // See if any elements differ.
  for (int i = 0; i < len; i++) {
    if (!element(i)->equal(l->element(i)))
      return false;
  }

  return true;
}

/////////////////////////////////////
// arithmetic

// monadic `+', absolute value
Slot List::qsort()
{
  List *m = (List*)mutate();
  m->Segment<Slot>::qsort();
  return m;
}

// monadic `-', negative absolute value
Slot List::reverse()
{
  List *m = (List*)mutate();
  m->Segment<Slot>::reverse();
  return m;
}

// dyadic `+', add
Slot List::add(const Slot &arg)
{
  return concat(arg);
}

#if 0
// dyadic `-', subtract
Slot List::subtract(const Slot &arg)
{
}

// dyadic `*', multiply
Slot List::multiply(const Slot &arg)
{
}

// dyadic '/', divide
Slot List::divide(const Slot &arg)
{
}

// dyadic '%', modulo
Slot List::modulo(const Slot &arg)
{
}
  
/////////////////////////////////////
// bitwise

// unary '~', invert
Slot List::invert()
{
}

// dyadic '&', bitwise and
Slot List::and(const Slot &arg)
{
}

// dyadic '^', bitwise xor
Slot List::xor(const Slot &arg)
{
}

// dyadic '|', bitwise or
Slot List::or(const Slot &arg)
{
}

// dyadic '<<', left shift
Slot List::lshift(const Slot &arg)
{
}

// dyadic '>>', right shift
Slot List::rshift(const Slot &arg)
{
}
#endif

/////////////////////////////////////
// sequence

// predicate - is object a sequence
bool List::isSequence()
{
  return true;
}

// length as sequence
int List::length() const
{
  return Vector<Slot>::Length();
}

// concatenate two sequences
Slot List::concat(const Slot &arg)
{
  if (!arg)
    return this;

  List *m = (List*)mutate();
  Slot other = arg->toSequence();
  ((Vector<Slot>*)m)->vconcat((List*)other);
  return m;
}

// dyadic `[]', range
Slot List::slice(const Slot &from, const Slot &len) const
{
  if (empty()) {
    if (((bool)len && (((int)len) != 0)) || ((int)from != 0)) {
      throw new Error("range", this, "Taking non-empty slice of an empty list");
    } else {
        return this;
    }
  }

  return new List(this, from, len);
}

// dyadic `[]', item
Slot List::slice(const Slot &from) const
{
  if (empty()) {
      throw new Error("range", this, "Taking non-empty slice of an empty list");
  }

  return element((int)from);
}

// replace subrange
Slot List::replace(const Slot &from, const Slot &len, const Slot &val)
{
   List *m = (List*)mutate();
   Slot other = val->toSequence();	// get the argument as a List
   m->Vector<Slot>::Replace((int)from,
                            len,
                            ((List*)other)->content(),
                            other->length());
   return m;
}

// replace item
Slot List::replace(const Slot &from, const Slot &val)
{
  // replace a slice with the value
  List *m = (List*)mutate();
  Slot other = val->toSequence();	// get the argument as a List
  m->Vector<Slot>::Replace((int)from,
                           1,
                           ((List*)other)->content(),
                           other->length());
  return m;
}

// replace subrange
Slot List::replace(const Slot &)
{
    return unimpl("monadic replace on List");
}

// insert
Slot List::insert(const Slot &from, const Slot &val)
{
  List *m = (List*)mutate();
  m->Vector<Slot>::vinsert((int)from, val->toSequence());
  return m;
}

// insert
Slot List::insert(const Slot &val)
{
  List *m = (List*)mutate();
  m->Vector<Slot>::vinsert(length(), val->toSequence());
  return m;
}

// delete subrange
Slot List::del(const Slot &from, const Slot &len)
{
  // assignment to a null length range is insertion or deletion
  List *m = (List*)mutate();
  m->Vector<Slot>::del(from, len);
  return m;
}

// delete item
Slot List::del(const Slot &from)
{
  List *m = (List*)mutate();
  m->Vector<Slot>::del(from, 1);
  return m;
}

// dyadic `in', matching subrange
Slot List::search(const Slot &search) const
{
  if (search->isSequence()) {
    Slot other = search->toSequence();
    Slot *found = Slot::search(content(),
                               ((List*)other)->content(),
                               other->length(),
                               Length());
    if (found)
      return found - content();
  } else {
    // singleton search
    Slot *found = Slot::search(content(), &search, 1, 0);
    if (found)
      return found - content();
  }
  return (Data*)0;	// NULL result
}

// return the implementation Segment
Slot List::toSequence() const
{
    return mutate();
    //was: new List(this, start, len); but I don't think it needs to be dup'd
}

// return an iterator
Slot List::iterator() const
{
    return clone();
    //return mutate();
  // was: return new List(this); but I don't think it needs to be dup'd
}

// List can act as an Iterator on List
bool List::More() const
{
  return truth()?1:0;
}

Slot List::Current()
{
  if (truth()) {
    return slice(0);
  } else {
    throw new Error("break", this, "Iterator exhaustion");
  }
}

Slot List::Next()
{
  if (truth()) {
    (*this)++;	// advance lower edge of Segment
  } else {
    throw new Error("break", this, "Iterator exhaustion");
  }
  return this;
}

Slot List::pop()
{
    if (length() < 1) {
        throw new Error("underflow", this, "stack underflow");
    }

    Slot r = (*this)[length() - 1];
    Vector<Slot>::del(-1, 1);
    return r;
}

void List::push(const Slot &val)
{
    //    cout << "push: stack before length is " << stack->Length() << "\n";
    vconcat(&val,1);
    //    cout << "push: stack after length is " << stack->Length() << "\n";
}

