// Slot - Coldstore smartpointers
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms

static char *id __attribute__((unused))="$Id: Slot.cc,v 1.4 1999/03/13 01:18:24 colin Exp $";

#include <typeinfo>

#include "Data.hh"
#include "Store.hh"

#include "String.hh"
#include "Tuple.hh"
#include "Error.hh"
#include "Integer.hh"
#include "BigInt.hh"
#include "Real.hh"

Slot::~Slot()
{
    if (!isEncInt() && data_)
        data_->dncount();
}

Slot::Slot(Data *d)
    : data_(d)
{
    DEBLOG(cerr << "Converting Data pointer " << d
           << " to Slot " << *this << '\n');
    if (!isEncInt() && d)
        d->upcount();
}

Slot::Slot(int l)
    : integer_(l<<1 | 1)	// encode the integer
{
    DEBLOG(cerr << "Converted int " << l
           << " to Slot " << *this << '\n');

    // we must check to see if l can be encoded
    if ((integer_ >> 1) != l) {
        // we're losing a bit of information ... don't encode
        data_ = Integer::getInteger(l);
    }
}

// reference constructor
Slot::Slot(Data &d)
    : data_(&d)
{
    DEBLOG(cerr << "Converted Data reference " << &d
           << " to Slot " << *this << '\n');
    if (!isEncInt() && &d)
        d.upcount();
}

// We're really cheating here - we take a const, but we cast it away.
Slot::Slot(Data const *d)
    : data_((Data*)d)
{
    DEBLOG(cerr << "Converted Data const pointer " << d
           << " to Slot " << *this << '\n');

    assert(!((int)d & 3) || !"Ill-formed Data *");
    if (d)
        d->upcount();
}

Slot::Slot(char const *str)
    : data_(new String(str))
{
    data_->upcount();
}

// copy constructor - dups reference
Slot::Slot(const Slot& f)
    : data_(f.data_)
{
    DEBLOG(cerr << "Duplicating Slot reference " << &f
           << " to Slot " << *this << '\n');

    if (!isEncInt() && data_)
        data_->upcount();
}

Slot::Slot(Slot &s)
    : data_(s.data_)
{
    DEBLOG(cerr << "Copying Slot " << s
           << " to Slot " << *this << '\n');
    if (!isEncInt() && data_)
        data_->upcount();
}

// assignment operator.  Initializes Slot
Slot &Slot::operator= (const Slot& f)
{
    // DO NOT CHANGE THE ORDER OF THESE STATEMENTS!
    // (This order properly handles self-assignment)
    if (!f.isEncInt() && f.data_) {
        f.data_->upcount();
    }

    // clean up the Data we own, if its refcount drops
    if (data_ && !isEncInt())
        data_->dncount();
    
    data_ = f.data_;
    return *this;
}

// assignment operator.  Initializes Slot
Slot& Slot::operator= (Data *d)
{
    assert(((int)d & 1) == 0 || !"Encoded Integer passed as Data*");

    // DO NOT CHANGE THE ORDER OF THESE STATEMENTS!
    // (This order properly handles self-assignment)
    if (d) {
        d->upcount();
    }

    // clean up the Data we own, if its refcount drops
    if (data_ && !isEncInt())
        data_->dncount();
    
    data_ = d;
    return *this;
}

// assignment operator.  Initializes Slot
Slot& Slot::operator= (int i)
{
    // clean up the Data we own, if its refcount drops
    if (data_ && !isEncInt())
        data_->dncount();
    
    integer_ = (i<<1) | 1;	// encode the integer
    
    // we must check to see if l can be encoded
    if ((integer_ >> 1) != i) {
        // we're losing a bit of information ... don't encode
        data_ = Integer::getInteger(i);
    }
    
    return *this;
}

// mkInteger - convert an encoded integer
//	into an instance of Integer
Data *Slot::mkInteger() const
{
    if (isEncInt()) {
        // we have to create an Integer from the pool of free Integers
        //data_ = Integer::getFree(integer_ >> 1);
        data_ = Integer::getInteger(integer_ >> 1);
    }
    
    return data_;
}

// dereference smartpointer
Slot::operator Data*() const
{
    return mkInteger();
}

Slot::operator Data&() const
{
    return *mkInteger();
}

Data *Slot::operator->() const
{
    return mkInteger();
}

Slot Slot::operator[](int index) const
{
    return mkInteger()->slice(index);
}

Slot Slot::operator[](Slot const &index) const
{
    return mkInteger()->slice(index);
}

Slot::operator UString *() const
{
    DEBLOG(cerr << "Converting Slot " << *this
           << " to UString " << data_ << '\n');

    if (isEncInt()
        || (!(data_->typeEQ(store->typeUString))
            && !(data_->typeEQ(store->typeString))
            && !(data_->typeEQ(store->typeBuiltin))
            ))
        throw new Error("type", *this, "expected UString");
    else
        return (UString*)data_;
}

Slot::operator String *(void) const
{
    DEBLOG(cerr << "Converting Slot " << *this
           << " to String " << data_ << '\n');

    if (isEncInt()
        || (!(data_->typeEQ(store->typeUString))
            && !(data_->typeEQ(store->typeString))
            && !(data_->typeEQ(store->typeBuiltin))
            ))
        throw new Error("type", *this, "expected String");
    else
        return (String*)data_;
}

bool Slot::isString() const
{
    if (!isEncInt()
        && (data_->typeEQ(store->typeUString)
            || data_->typeEQ(store->typeString)
            || data_->typeEQ(store->typeBuiltin)
            ))
        return true;
    else
        return false;
}

Slot::operator List *() const
{
    if (isEncInt() || !(data_->typeEQ(store->typeList)))
        throw new Error("type", *this, "expected List");
    else
        return (List*)data_;
}

Slot::operator NList *() const
{
    if (isEncInt() || !(data_->typeEQ(store->typeNList)))
        throw new Error("type", *this, "expected NList");
    else
        return (NList*)data_;
}

Slot::operator Tuple *() const
{
    if (isEncInt() || !(data_->typeEQ(store->typeTuple)))
        throw new Error("type", *this, "expected Tuple");
    else
        return (Tuple*)data_;
}

bool Slot::isList () const
{
    if (!isEncInt()
        && (
            (data_->typeEQ(store->typeList))
            ||(data_->typeEQ(store->typeNList))
            ||(data_->typeEQ(store->typeTuple))
            )
        )
        return true;
    else
        return false;
}

Slot::operator Builtin *(void) const
{
    if (isEncInt() || !(data_->typeEQ(store->typeBuiltin)))
        throw new Error("type", *this, "expected Builtin");
    else
        return (Builtin*)data_;
}

Slot::operator Dict *(void) const
{
    if (isEncInt() || !(data_->typeEQ(store->typeDict)))
        throw new Error("type", *this, "expected Dict");
    else
        return (Dict*)data_;
}

Slot::operator DictIt *(void) const
{
    if (isEncInt() || !(data_->typeEQ(store->typeDictIt)))
        throw new Error("type", *this, "expected DictIt");
    else
        return (DictIt*)data_;
}

Slot::operator Symbol *(void) const
{
    if (isEncInt() || !(data_->typeEQ(store->typeSymbol)))
        throw new Error("type", *this, "expected Symbol");
    else
        return (Symbol*)data_;
}

Slot::operator Namespace *(void) const
{
    if (isEncInt() || !(data_->typeEQ(store->typeNamespace)))
        throw new Error("type", *this, "expected Namespace");
    else
        return (Namespace*)data_;
}

Slot::operator Error *(void) const
{
    if (isEncInt() || !(data_->typeEQ(store->typeError)))
        throw new Error("type", *this, "expected Error");
    else
        return (Error*)data_;
}

Slot::operator Integer *() const
{
    if (isEncInt()) {
        return (Integer*)mkInteger();
    } else if (data_->typeEQ(store->typeInteger)) {
        return (Integer*)data_;
    } else if (data_->typeEQ(store->typeBigInt)) {
        return ((BigInt*)data_)->asInteger();
    } else if (data_->typeEQ(store->typeReal)) {
        return ((Real*)data_)->asInteger();
    } else
        throw new Error("type", *this, "expected Integer");
}

Slot::operator BigInt *() const
{
    if (isEncInt()) {
        return new BigInt(integer_ >> 1);
    } else if (data_->typeEQ(store->typeInteger)) {
        return new BigInt(((Integer*)data_)->value());
    } else if (data_->typeEQ(store->typeReal)) {
        return new BigInt(((Real*)data_)->asGMP());
    } else if (data_->typeEQ(store->typeBigInt)) {
        return (BigInt*)data_;
    } else
        throw new Error("type", *this, "expected Integer");
}

Slot::operator Real *() const
{
    if (isEncInt()) {
        return new Real(integer_ >> 1);
    } else if (data_->typeEQ(store->typeInteger)) {
        return new Real((Integer*)data_);
    } else if (data_->typeEQ(store->typeBigInt)) {
        return new Real(((BigInt*)data_)->asGMP());
    } else if (data_->typeEQ(store->typeReal)) {
        return (Real*)data_;
    } else
        throw new Error("type", *this, "expected Integer");
}

bool Slot::isNumeric() const
{
    if (isEncInt()
        || (data_->typeEQ(store->typeInteger))
        || (data_->typeEQ(store->typeBigInt))
        || (data_->typeEQ(store->typeReal)))
        return true;
    else
        return false;
}

Slot::operator char *(void) const
{
    char *result = (char *)(((String*)*this)->content());
    DEBLOG(cerr << "Converted Slot " << *this
           << " to char* " << result << '\n');
    return result;
}

Slot::operator int() const {
    DEBLOG(cerr << "Converting Slot " << *this << " to int\n");
    if (isEncInt()) {
        DEBLOG(cerr << "Converted Slot " << *this
               << " to int " << (integer_ >> 1) << '\n');
        return integer_>>1;
    } else if (!data_) {
        throw new Error("type", (Data*)NULL, "expected integer");
    } else if ((data_->typeEQ(store->typeBigInt))) {
        return ((BigInt*)data_)->asint();
    } else if ((data_->typeEQ(store->typeReal))) {
        return ((Real*)data_)->asint();
    } else if ((data_->typeEQ(store->typeInteger))) {
        return ((Integer*)data_)->value();
    } else {
        throw new Error("type", *this, "expected integer");
    }
}

Slot::operator long() const {
    DEBLOG(cerr << "Converting Slot " << *this << " to int\n");
    if (isEncInt()) {
        DEBLOG(cerr << "Converted Slot " << *this
               << " to int " << (integer_ >> 1) << '\n');
        return integer_>>1;
    } else if (!data_) {
        throw new Error("type", (Data*)NULL, "expected integer");
    } else if ((data_->typeEQ(store->typeBigInt))) {
        return ((BigInt*)data_)->asint();
    } else if ((data_->typeEQ(store->typeReal))) {
        return ((Real*)data_)->asint();
    } else if ((data_->typeEQ(store->typeInteger))) {
        return ((Integer*)data_)->value();
    } else {
        throw new Error("type", *this, "expected integer");
    }
}

const char *Slot::typeId() const {
    if (isEncInt())
        return "7Integer";
    else
        return typeid(*data_).name();
}

#if 0
// get the real pointer to data (accounting for any Forward'ing)
Data *Slot::forward()
{
    assert(_data);	// can't dereference a NULL Slot
    
    if ((Type)*_data == isForward) {
        // _data in fact points to a Forward
        // so we update our data_ to point to the forwardee
        *this = ((Forward *)_data)->moved();	// update our pointer
    }
    
    return _data;	// return the actual data, stripped of forwardings
}

// A Data method needs to change something in itself
// Thus it first checks if this is the only pointer to *data_
Data *Slot::mutate() const
{
    //forward();
    if (!isEncInt()) {
        // if the Data is immutable,
        // make a clone and point-at/return it
        data_ = data_->mutate();
    }

    // assert((integer & 1) || data_->refs() == 1);	// clearly, to be mutable, refs == 1
    return *this;	// return the (possibly) new Data to method:
}
#endif

// find offset of leftmost NULL element in range 
int Slot::findNull(const Slot *from)
{
    const Slot *p = from;
    while (*p);
    return p - from;
}

// consistency checker for Slot
void Slot::check(const Slot *from, size_t len)
{
    for (size_t i = 0; i < len; i++) {
        if (from[i].integer_ && !(from[i].isEncInt()))
            ((Data*)(from[i]))->check();
    }
}

// duplicate the range
// perform necessary reference counting
Slot *Slot::dup(Slot *to, const Slot *from, size_t range)
{
    for (size_t i = 0; i < range; i++) {
        to[i] = from[i];
    }
    return to;
}

// find a single matching T
Slot *Slot::find(Slot *from, const Slot &_c, size_t range)
{
    for (size_t i = 0; i < range; i++) {
        if (from[i] == _c)
            return from + i;
    }
    return NULL;
}

ostream &Slot::Dump(ostream& out) const
{
    if (integer_ != 0) {
        mkInteger();
        out << data_->typeId();
        out.form("[0x%08x]/", data_);
        data_->dump(out);
    } else {
        out << "NULL";
    }
    return out;
}

// Data ordering over Slot
int Slot::order(const Slot& d) const
{
    if (integer_ - d.integer_) {
        // they're not numerically equal - point to different things
        if (isEncInt()) {
            // encoded Integer - temporarily decode it and recurse
            Slot temp(Integer::getInteger(integer_>>1));
            return temp.order(d);			// recurse
        } else {
            // this is not an encoded Integer - use the virtual comparison
            if (integer_) {
                // this is non-null
                if (d.integer_)
                    return data_->order(d);	// try comparison on its merits
                else
                    return 1;		// other is null, must be greater
            } else {
                // this is null
                if (d.integer_)
                    return -1;		// this is null, so d must be greater
                else
                    return 0;		// they're both null
            }
        }
    } else {
        // They're the same entity, ergo ==
        return 0;
    }
}

int Slot::order(const Slot& d1, const Slot& d2)
{
    return d1.order(d2);
}

// Data equality over Slot
bool Slot::equal(const Slot& d) const
{
    if (integer_ != d.integer_) {
        // they're not numerically equal - point to different things
        if (isEncInt()) {
            // this is an encoded Integer - temporarily decode it and recurse
            Slot temp(Integer::getInteger(integer_>>1));
            return temp.equal(d);			// recurse
        } else {
            // this is not an encoded Integer - use the virtual comparison
            if (integer_) {
                // this is non-null
                if (d.integer_)
                    return data_->equal(d);	// try equality on its merits
                else
                    return false;		// other is null, this is not
            } else {
                // this is null
                if (d.integer_)
                    return false;		// this is null, other isn't
                else
                    return true;		// both null
            }
        }
    } else {
        // They're the same entity, ergo ==
        return true;
    }
}

void Slot::segFill(Slot *to, Slot const &filler, int count)
{
    Slot *sptr = to;
    for (int i=count;  i; i--)
        *(sptr++) = filler;
}

// search - find a needle in this haystack O(n^2)
Slot *Slot::search(Slot *haystack, const Slot *needle,
                   size_t len, size_t range)
{
    for (size_t i = 0; i < (range - len + 1); i++) {
        size_t j;
        for (j = 0; j < len; j++) {
            if (*(haystack + i + j) != *(needle + j))
                break;
        }
        if (j == len)
            return haystack + i;
    }
    return NULL;
}

ostream& operator<< (ostream& out, const Slot & slot)
{
    out << &slot << " [" << slot.data_ << "] ";
    if (slot.data_) {
        out << slot.typeId();
    } else {
        out << "NIL";
    }
    //  out << slot.typeId() << '(' << slot->toconstruct() << ')';
    return out;
}

#if MAIN
#include <iostream.h>
#include <typeinfo>

main()
{
    cerr << typeid(Slot).name() << " size: " << sizeof(Slot) << '\n';
}
// reports 4 bytes
#endif 
