// Startup - Server Initialization
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms

static char *id __attribute__((unused))="$Id$";

#include <iostream.h>
#include <assert.h>
#include <fcntl.h>

#include <elf.hh>
#include "Data.hh"
#include "Store.hh"
#include "Task.hh"

#include "Vector.hh"
#include "String.hh"
#include "Tuple.hh"
#include "Error.hh"
#include "Builtin.hh"
#include "Dict.hh"
#include "Symbol.hh"
#include "List.hh"
#include "Integer.hh"
#include "BigInt.hh"

extern void shutdown();	// the default shutdown function

/** Create a bootstrap symbol table of Builtins for this Elf
 */
static Builtin **traverse(Elf *elf, Csym *sym, int cnt = 0)
{
    Builtin **table = NULL;
    if (sym != elf->symQ()) {
        table = traverse(elf, (Csym*)(sym->next()), cnt + 1);
        table[cnt] = new Builtin(sym);
        table[cnt]->upcount();

        DEBLOG(cerr << cnt
               << " traverse " << sym->symname()
               << " at: " << sym
               << " --> " << table[cnt]
               << '\n');
    } else {
        // this is the last in Q.
        table = (Builtin**)malloc(sizeof(Builtin*) * (1 + cnt));
        table[cnt] = 0;
        table[cnt-1] = new Builtin(sym);
        table[cnt-1]->upcount();
    }
    return table;
}

/** register any new Builtins in the bootstrap symbol table.
 */
static void registerElf(Builtin **table)
{
    for (Builtin **bpp = table; *bpp; bpp++) {
        try {
            // look for a pre-existing Builtin
            if (!store->Registered->search(*bpp)) {
                store->Registered->insert((*bpp)->sym->symname(), *bpp);
            }
        } catch (Error *e) {
            // New Builtin
            if (((Symbol*)*e) == store->Errors->search("dictnf")) {
                // register new Builtin
                store->Registered->replace((*bpp)->sym->symname(), *bpp);
            } else {
                // some other error
                throw e;
            }
        }
        (*bpp)->dncount();	// discard if not new

        // Note: we don't replace existing registered Builtins.
        // since Csym can't be destroyed, any Builtin (once created)
        // will reference an existing Csym, and so is valid forever

        // Only way to reclaim a registered Builtin is by rebuilding
        // the store
    }
}

/** called to register Csyms as Builtins for a foreign module
 */
void RegisterElf(Elf *elf)
{
    cerr << "RegisterElf " << elf << "\n";
    if (!elf->symQ()) {
        DEBLOG(cerr << elf << "empty\n");
        return;
    }
    Builtin **table = traverse(elf, (Csym*)(elf->symQ()->next()));
    registerElf(table);
    free(table);
    cerr << "RegisterElf " << elf << " done\n";
}

/** a utility to dump a Csym
 */
void dump_csym(Elf *elf, Csym *sym)
{
    cerr << "dump " << sym->symname()
         << "at: " << sym
         << " next: " << sym->next()
         << " in: " << sym->in
         << " elf: " << elf
         << " symQ " << sym->in->symQ()
         << '\n';
}

/** find the Builtin/VTBL implementing this class in a bootstrap symbol table
 */
static Builtin *findV(Builtin **table, char *class_name, bool chuck=true)
{
    char name[256];
    sprintf(name, "__vt_%d%s", strlen(class_name), class_name);
    for (Builtin **bpp = table; *bpp; bpp++) {
        DEBLOG(cerr << "findV " << name << ", " << **bpp << '\n');
        if (!strcmp(name, (char *)**bpp)) {
            (*bpp)->upcount();	// maintain this Builtin if only for Store
            DEBLOG(cerr << (*bpp)->dump(cerr) << " FOUND\n");
            return *bpp;
        }
    }
    if (chuck) {
        sprintf(name, "Can't find vtbl for %s", class_name);
        throw name;
    }
    return NULL;
}

// initialize Errors namespace with the errors we throw
// note: this can be overriden in the executable,
// just make sure it's got at least these and ends in NULL
char *ColdErrors[] = {
    "dictnf",
    "type",
    "args",
    "break",
    "builtin",
    "conformant",
    "exhausted",
    "illegal",
    "nil",
    "parse",
    "range",
    "symbol",
    "unimplemented",
    "value",
    "underflow",
    NULL
};

// init - initialize subsystems
Store::Store(Builtin **table)
{
    // perform once-only initializations

    // Init well-known DataType Managers
    typeString = findV(table, "String");	// dynamic strings
    typeBuiltin = findV(table, "Builtin");	// callable C functions
    typeDict = findV(table, "Dict");		// ordered mapping
    typeDictIt = findV(table, "DictIt");	// Dict iterator 
    typeError = findV(table, "Error");		// error over Tuple
    typeInteger = findV(table, "Integer");	// 4 byte signed longs
    typeList = findV(table, "List");		// dynamic lists
    typeNList = findV(table, "NList");		// dynamic n-ary lists
    typeNamespace = findV(table, "Namespace");	// Dict of Symbols
    typeUString = findV(table, "UString");	// caseless strings
    typeSymbol = findV(table, "Symbol");	// interned strings
    typeTuple = findV(table, "Tuple");		// static lists
    typeBigInt = findV(table, "BigInt");	// GMP integers
    typeReal = findV(table, "Real");		// GMP floats
    
    // Errors namespace: the errors coldstore throws
    Errors = new Namespace();
    Errors->upcount();		// Must protect Namespace store-Errors

    // Registered Builtin namespace (aka symbol table)
    Registered = new Namespace();
    Registered->upcount();	// Must protect Namespace store-Registered

    // Coldstore global Namespace
    Names = new Namespace();
    Names->upcount();		// Must protect Namespace store-Names

    initialized = true;
}

void cold_process()
{
    cerr << "Called ColdStore default process\n";
}

/** called when Elf has loaded this module
 */
void initElf(Elf *elf)
{
#ifdef DEBUGLOG
    cerr << "Store's initElf called\n";
    elf->forallSym(dump_csym);
#endif
    try {

        // create type mapping to Csym::VTBL syms
        if (!store) {
            if (!elf->symQ()) {
                cerr << elf << "empty\n";
                return;
            }
            Builtin **table = traverse(elf, (Csym*)(elf->symQ()->next()));
            store = new Store(table);
            
            // initialize Errors namespace with the errors we throw
            for (char **errs = ColdErrors; *errs; errs++) {
                //const Slot err(new String(*errs));
                String *err = new String(*errs);
                store->Errors->insert(err, err); //mkSym
            }
#ifdef DEBUGLOG
            store->Errors->dump(cerr);
#endif
            // Register all the Builtins
            registerElf(table);
        } else {
            // Register any new Builtins
            RegisterElf(elf);
        }
    } catch (Error *e) {
        e->dump(cerr) << '\n';
    }

    // replace gmp's allocation functions
    mp_set_memory_functions(BigInt::GmpAlloc,
                            BigInt::GmpReAlloc,
                            BigInt::GmpDeAlloc);
#ifdef DEBUGLOG
    cerr << "ColdStore initElf() complete\n";
#endif
    return;

    try {
        //cerr << init() << '\n';
        cold_process();
    }
    catch (Error *e) {
        cerr << "\nUncaught Error: ";
        ((Data*)e)->dump(cerr);
        cerr << '\n';
    }

    shutdown();
}

