// Symbol.cc - global symbols
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms

static char *id __attribute__((unused))= "$Id: Symbol.cc,v 1.3 1999/03/13 00:40:30 colin Exp $";

#include "Data.hh"
#include "Store.hh"

#include "String.hh"
#include "Tuple.hh"
#include "Error.hh"
#include "Dict.hh"
#include "List.hh"
#include "Symbol.hh"


Symbol::Symbol(const Slot &name)
  : Tuple(-3),
    _namespace((Data*)0),
    _name(name),
    _value((Data*)0)
{
    throw new Error("illegal", this, "Symbol without a Namespace");
    DEBLOG(dump(cerr) << " New Symbol\n");
}

Symbol::Symbol(const Slot &name, const Namespace *nspace)
  : Tuple(-3),
    _namespace(nspace),
    _name(name),
    _value((Data*)0)
{
    DEBLOG(dump(cerr) << "New Symbol\n");
}

Symbol::Symbol(const Slot &name, const Namespace *nspace, const Slot &value)
  : Tuple(-3),
    _namespace(nspace),
    _name(name),
    _value(value)
{
    DEBLOG(dump(cerr) << "New Symbol\n");
}

Symbol::Symbol(Symbol &sym)
  : Tuple(-3),
    _namespace(),
    _name(),
    _value()
{
  throw new Error("illegal", &sym, "Copying a Symbol");
}

Symbol::~Symbol()
{
    DEBLOG(dump(cerr) << " Destroy Symbol\n");
    // Clean up the Namespace's reference to this
    ((Data*)space())->del(this);
}

// Uses the copy constructor
Data *Symbol::clone(void *where)const
{
  throw new Error("symbol", this, "An attempt was made to duplicate a Symbol");
  return (Data*)0;	// Symbols are unique
}

// returns a mutable copy of this
Data *Symbol::mutate(void *where) const
{
  throw new Error("symbol", this, "An attempt was made to mutate a Symbol");
  return (Data*)0;	// Symbols are unique
}

ostream &Symbol::dump(ostream& out) const
{
    out << typeId() << ':';
    _name.Dump(out);
    out << "=";
    _value.Dump(out);
    return out;
}

// Here's the big payoff - comparison is numeric
int Symbol::order(Slot const &arg) const
{
  return typeOrder(arg) || (this - (Symbol*)arg);
}

// Here's the big payoff - comparison is numeric
bool Symbol::equal(Slot const &arg) const
{
  return (Data*)this == (Data*)arg;
}

// Symbol assign - note this can be overridden
Slot Symbol::replace(const Slot &value)
{
  _value = value;
  return value;
}

// replace subrange
Slot Symbol::replace(const Slot &, const Slot &)
{
    return unimpl("dyadic replace on Symbol");
}
Slot Symbol::replace(const Slot &, const Slot &, const Slot &)
{
    return unimpl("triadic replace on Symbol");
}

// Local

// return Namespace of Symbol
Namespace *Symbol::space()
{
  if (_namespace)
    return _namespace;
  else
    return store->Names;
}

Slot Symbol::operator=(const Slot &value)
{
    replace(value);
    return value;
}

//////////////////////
/// Namespace

Namespace::Namespace() {}

Namespace::Namespace(const Slot &init)
{
    int l = init->length();
    for (int i = 0; i < l; i+=2) {
        insert(init[i], init[i+1]);
    }
}

Namespace::Namespace(Namespace &n)
{
    throw Error("illegal", n, "Can't copy construct a Namespace");
}

Namespace::~Namespace() {}

Data *Namespace::clone(void *where) const
{
    return (Data*)this;	// namespaces are inherently mutable
}

Data *Namespace::mutate(void *where) const
{
    DEBLOG(dump(cerr) << " mutate\n");
    return (Data*)this;
}

// replace association
// In a Namespace, assignment has the effect of assigning a value to a Symbol
// If the Symbol doesn't exist, it is created.  Only Namespace can construct 
// Symbols.
Slot Namespace::replace(const Slot &from, const Slot &value)
{
    DEBLOG(dump(cerr)
           << " replace " << from->dump(cerr)
           << " with " << value->dump(cerr)
           << '\n');
  // a single item
    SlotPair *found = Find(from);
    if (!found){
        // unknown symbol - throw
        throw new Error("dictnf", from, "Symbol not found");
    } else {
        // bind new value
        Replace(found, value);
    }
    return value;
}

// insert association
// In a Namespace, assignment has the effect of assigning a value to a Symbol
// If the Symbol doesn't exist, it is created.  Only Namespace can construct 
// Symbols.
Slot Namespace::insert(const Slot &key, const Slot &value)
{
    DEBLOG(dump(cerr)
           << " insert " << key->dump(cerr)
           << " with " << value->dump(cerr)
           << '\n');
    // a single item
    Data *found = search(key);
    if (!found){
        // unknown symbol - create and add it
        Slot s = new (this) Symbol(key, this, value);
        Add(key, s);
        return s;
    } else {
        // bind new value
        return ((Symbol*)found)->replace(value);
    }
}

// insert association
// In a Namespace, assignment has the effect of assigning a value to a Symbol
// If the Symbol doesn't exist, it is created.  Only Namespace can construct 
// Symbols.
Slot Namespace::insert(const Slot &key)
{
    DEBLOG(dump(cerr)
           << " insert " << key->dump(cerr)
           << '\n');
    // a single item
    Data *found = search(key);
    if (!found){
        // unknown symbol - create and add it
        Slot s = new (this) Symbol(key, this);
        Add(key, s);
        return s;
    } else {
        // bind new value
        return found;
    }
}

// replace subrange
Slot Namespace::replace(const Slot &)
{
    return unimpl("monadic replace on Namespace");
}
Slot Namespace::replace(const Slot &, const Slot &, const Slot &)
{
    return unimpl("triadic replace on Namespace");
}

// delete subrange
Slot Namespace::del(const Slot &from)
{
    // TODO? put some permissions stuff here to stop people deleting bindings
    // of Symbols with non-zero external references.
    DEBLOG(dump(cerr)
           << " delete " << from->dump(cerr));
    if (from->typeEQ(store->typeSymbol)) {
        Slot retval = slice(from[1]);
        Dict::del(from[1]);	// delete using name property of Symbol
        return retval;
    } else {
        Slot retval = slice(from);
        Dict::del(from);
        return retval;
    }
}

// delete subrange
Slot Namespace::del(const Slot &, const Slot &)
{
    return unimpl("dyadic del on Namespace");    
}
