// Symbol - global name/value binding
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms
// $Id: Symbol.hh,v 1.2 1999/03/11 06:51:20 colin Exp $

#ifndef SYMBOL_HH
#define SYMBOL_HH

class Namespace;

/** A carrier for associations within a @ref Namespace
 *
 * Symbol is the carrier for associations between name and value within
 * a @ref Namespace.
 *
 * Symbols are unique within a @ref Namespace.  In effect, a Symbol's name
 * is a (Namspace*,Slot) pair.
 *
 * Symbols compare by address and not by value.
 * They are therefore an excellent choice for storage in @ref Hash tables
 * and such.
 *
 * Symbols may only be constructed by a Namespace.
 */
class Symbol
    : public Tuple
{
    // these are disallowed - will abort

    /** disallowed: default constructor
     *
     * Symbols can only exist in the context of a @ref Namespace
     * and need a name, so it makes no sense to create a default Symbol
     */
    Symbol();

    /** disallowed: coldstore constructor
     *
     * Symbols can only exist in the context of a @ref Namespace
     * so it makes no sense to create a Symbol without a @ref Namespace
     */
    Symbol(const Slot &name);

    /** disallowed: copy constructor
     *
     * Symbols are unique within a @ref Namespace
     * so it makes no sense to copy a Symbol
     */
    Symbol(Symbol &sym);

    /** disallowed - cloning a symbol
     * Symbols are unique within a @ref Namespace
     * so it makes no sense to copy a Symbol
     */
    virtual Data *clone(void *where = (void*)0)const;
    
    /** disallowed - mutating a symbol
     * Symbols are inherently mutable,
     * and only mutable from @ref Namespace
     * that this has been called suggests a protocol violation
     */
    virtual Data *mutate(void *where = (void*)0) const;

    /** disallowed: it's meaningless
     */
    virtual Slot replace(const Slot &, const Slot &, const Slot &);

    /** disallowed: it's meaningless
     */
    virtual Slot replace(const Slot &, const Slot &);

protected:
    
    // These values are accessible via Tuple

    /** @ref Namespace within which this is defined */
    Slot _namespace;
    /** name of Symbol (can be any type, usually a string */
    Slot _name;
    /** Symbol's value (any type of Data) */
    Slot _value;
    
    // Symbols can only be constructed by a Namespace
    friend class Namespace;

    /** create a NULL-valued Symbol in a @ref Namespace
     * @param name Symbol's name
     * @param nspace Symbol's namespace
     */
    Symbol(const Slot &name, const Namespace *nspace);

    /** create a Symbol in a @ref Namespace
     * @param name Symbol's name
     * @param nspace Symbol's namespace
     * @param value Symbol's value
     */
    Symbol(const Slot &name, const Namespace *nspace, const Slot &value);

    /** destroy a Symbol
     */
    virtual ~Symbol();
    
public:
    
    /** dump the object to a stream
     * @param ostream stream onto which to dump the object
     */
    virtual ostream &dump(ostream&) const;
    
    /** order two Symbols
     *
     * Symbols are always interned within some @ref Namespace.
     * This means they're unique, and immobile.
     *
     * Therefore they can be compared on the
     * basis of their address, not their content.  This is fast.
     *
     * @param arg Symbol to be ordered relative to this
     * @return 1,0,-1 depending on object order
     */
    virtual int order(const Slot &arg) const;

    /** compare two Symbols for strong equality
     *
     * Symbols are always interned within some @ref Namespace.
     * This means they're unique, and immobile.
     *
     * Therefore means they can be compared for strong equality on the
     * basis of their address, not their content.  This is fast.
     *
     * @param arg Symbol to be ordered relative to this
     * @return 1,0,-1 depending on object order
     */
    virtual bool equal(const Slot &arg) const;
    
    /** assign to Symbol
     * @param value value to be assigned
     * @return a copy of this with the assignment performed
     */
    virtual Slot replace(const Slot &value);
    
    /** the namespace within which this Symbol is defined
     * @returns defining Namespace
     */
    Namespace *space();

    /** this Symbol's name
     * convenience accessor
     * @returns Symbol's name
     */
    Slot name() {return _name;}

    /** this Symbol's value
     * convenience accessor
     * @returns Symbol's value
     */
    Slot value() {return _value;}
    
    /** Assign a value to this Symbol
     * convenience mutator - equivalent to @ref Symbol::replace()
     * @returns Symbol's new value
     */
    Slot operator=(const Slot &vale);
};

/** associations between @ref Symbol names and Symbols
 */
class Namespace
    : public Dict
{
    /** disallowed:
     * it makes very little sense to consider a range of Symbols.
     *
     * Having said that, it could be used to clear a group of Symbols
     * I've seen no compelling use for it, so at the moment it's out.
     */
    virtual Slot replace(const Slot &, const Slot &, const Slot&);

    /** disallowed:
     * it makes very little sense to NULL out a @ref Symbol,
     * which is what this would seem to do.
     */
    virtual Slot replace(const Slot &);

    /** disallowed:
     * I don't really want to allow the deletion of a range of Symbols,
     * it just seems too powerful.
     */
    virtual Slot del(const Slot &, const Slot &);

    /** disallowed: copy connstruction of a Namespace */
    Namespace(Namespace &n); 

    /** clone this Namespace
     * Namespaces are inherently mutable, and their contents 
     * are specific to the Namespace,
     * so it makes no sense to clone one unless to duplicate all
     * Symbols ... which's a lot of work to no obvious purpose.
     */
    virtual Data *clone(void *where = (void*)0)const;

public:
    /** construction of a Namespace */
    Namespace(); 

    /** construction of a Namespace from an association list, or map
     *
     * Namespaces can be constructed from a list of name,value pairs.
     */
    Namespace(const Slot &init); 

    /** destruction of a Namespace
     *
     * this will inevitably lead to the destruction of all Symbols
     * in the Namespace (won't it?)
     */
    ~Namespace();

    /** make this mutable
     * Namespaces are inherently mutable anyway.
     * @returns an identity - this.
     */
    virtual Data *mutate(void *where = (void*)0) const;

    /** assign a Symbol's value in this Namespace
     *
     * Throws if the Symbol's not already defined
     * @param key the Symbol's name
     * @param value the Symbol's value
     * @returns the Symbol's new value
     * @exception ~dictnf
     */
    virtual Slot replace(const Slot &key, const Slot &value);

    /** create/assign a Symbol's value in this Namespace
     *
     * Creates a new Symbol with the given value, or assigns the
     * value to an existing Symbol.
     * @param key the Symbol's name
     * @param value the Symbol's value
     * @returns the Symbol's new value
     */
    virtual Slot insert(const Slot &key, const Slot &value);

    /** create a Symbol in this Namespace
     *
     * Creates a new Symbol with the given value.
     * Does not change the value of an existing Symbol.
     * @param key the Symbol's name
     * @returns the Symbol's value
     */
    virtual Slot insert(const Slot &key);

    /** delete a Symbol from this Namespace
     *
     * Symbol must exist, or ~dictnf is thrown
     * @param from Symbol to delete
     * @returns Symbol's value
     * @exception ~dictnf
     */
    virtual Slot del(const Slot &from);
};


#endif
