// TupleBase - templated class for a static array of elements of arbitrary type
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms
// 	$Id: TupleBase.hh,v 1.1.1.1 1999/01/14 04:25:56 colin Exp $	

#ifndef TUPLEBASE_HH
#define TUPLEBASE_HH

#include <typeinfo>
#include <stdlib.h>
#include <stddef.h>
#include <iostream.h>
#include <assert.h>

#include "Vector.hh"

class ostream;

/** a templated arbitrary- but fixed- size collection of some basetype

    TupleBase is implemented as an immobile allocation which contains by
    inclusion, not reference.

    Note: C++ Children of TupleBase have their internal class data members
    exposed to indexing by the _content[0] array data member.
*/
template <class T>
class TupleBase
  : public Data
{

protected:

  virtual ~TupleBase() {
    check();
    T::destroy(_content, _length);
  }

protected:
    /** dump Tuple to file
     */
    friend ostream& operator<< <T> (ostream& out, const TupleBase<T> & vec);
    
    /** length of Tuple
     */
    int _length;
    
    /** Tuple's contents (overlayed on whatever follows)
     */
    mutable T _content[];
    
public:
    
    /////////////////////////////
    // Simple constructors
    /** construct a Tuple with pre-allocation of storage
     */
    TupleBase(int size);

    /** consistency check
     */
    virtual void check(int i = 1) const;


    /** allocate space for Tuple of given byte size, with extra overhead
     */
    void *operator new (unsigned int size, void *allocator = (void*)0) {
        return Data::operator new(size, allocator);
    }

    /** allocate space for Tuple of given size, with extra overhead
     */
    void *operator new (unsigned int size,
                        int extra, void *allocator = (void*)0) {
        return operator new(size + (sizeof(T)*extra), allocator);
    }

    /////////////////////////////////////
    // copy constructors of sundry kinds

    /** Copy T* array of a given size
     */
    TupleBase(const T *contentT, int size);

    /** Copy null-terminated T* array
     */
    TupleBase(const T *contentT);

    /** Copy Segment
     */
    TupleBase(const Segment<T> &contentT, int size=0);

    /** Copy Vector
     */
    TupleBase(const Vector<T> *contentT, int size = 0);

    /** Repeat T& as array
     */
    TupleBase(const T &contentT, int repetition = 1);

    /** Copy a Tuple<T>*'s contents
     */
    TupleBase(const TupleBase<T> *contentT);

    /** Copy a Tuple<T>&'s contents
     */
    TupleBase(const TupleBase<T> &contentT);

    /** Copy subrange Tuple<T>*
     */
    TupleBase(const TupleBase<T> *contentT, int _start, int l);

    /** Copy subrange Tuple<T>&
     */
    TupleBase(const TupleBase<T> &contentT, int _start, int l);

    /** union - v1 + v2
     */
    TupleBase(const TupleBase<T> *v1, const TupleBase<T> *v2);

    /** union - v1 + v2
     */
    TupleBase(const TupleBase<T> &v1, const TupleBase<T> &v2);

    /** argument range modifier
     */
    int endRelative(int where) const {
        // negative where is end-relative
        return (where < 0)? _length + where: where;
    }

public:
    T *content() const {
        return &(_content[0]);
    }

    int Length() const {
        return _length;
    }

    T &operator[](int offset) const {	// array accessor
        assert(offset >= 0);
        return _content[offset];
    }

    T &element(int offset) {	// array accessor
        assert(offset >= 0);
        return _content[offset];
    }

    const T &const_element(int offset) const {	// array accessor
        assert((offset >= 0) && (offset < _length));
        return _content[offset];
    }

    operator T*() const {	// array accessor
        return _content;
    }

    bool encloses(const Segment<T> &subrange) {
        return (subrange.content() >= _content)
            && ((_content + _length)
                >= (subrange.content() + subrange.Length()));
    }

    const T *last() const {	// ref to last in Segment
        return _content + _length - 1;
    }

    int bounds() const {
        return (Memory::mySize() - offsetof(TupleBase<T>, _content))
            / sizeof(T);
    }

    // replace range with array
    TupleBase<T> *Replace(int where, int what, const T *addT, int additional);
    inline TupleBase<T> *Replace(int where, int what, const T *addT) {
        return Replace(where, what, addT, T::findNull(addT));
    }
    
    inline TupleBase<T> *Replace(int where, int what, const T &addT,
                                 int repeat=1) {
        return Replace(where, what, &addT, repeat);
    }
    inline TupleBase<T> *Replace(int where, int what, const Segment<T> *addT) {
        return Replace(where, what, addT->content(), addT->Length());
    }
    inline TupleBase<T> *Replace(int where, int what,
                                 const TupleBase<T> &addT) {
        return Replace(where, what, addT.content(), addT.Length());
    }

    // ColdStore interface
    virtual Data *clone(void *where) const {	// Uses the copy constructor
        return new (where) TupleBase<T>(this);
    }

    virtual Data *mutate(void *where) const {	// return mutable copy of this
        return clone(where);
    }

    /** throw an error - shouldn't be called
     */
    virtual int order(Slot const &) const {
        unimpl("operator order undefined on TupleBase<>");
        return 0;
    }
};

// Construct a TupleBase of given size
template <class T>
TupleBase<T>::TupleBase(int size)
    : _length(size)
{
    T::init(_content, _length);
    check();
}

// Copy T* array of a given size
template <class T>
TupleBase<T>::TupleBase(const T *contentT, int size)
    : _length(size)
{
    assert(size > 0);
    T::check(contentT, size);
    T::init(_content, _length);
    T::dup(_content, contentT, _length);
    check();
}

// Copy a Segment
template <class T>
TupleBase<T>::TupleBase(const Segment<T> &contentT, int size)
    : _length(size?size:contentT.Length())
{
    assert(size > 0);
    T::check(contentT.content(), contentT.Length());
    T::init(_content, _length);
    T::dup(_content, contentT.content(), _length <? contentT.Length());
    check();
}

// Copy a Segment
template <class T>
TupleBase<T>::TupleBase(const Vector<T> *contentT, int size)
    : _length(size?size:contentT->Length())
{
    assert(size > 0);
    contentT->check();
    T::check(contentT->content(), contentT->Length());
    T::init(_content, _length);
    T::dup(_content, contentT->content(), _length <? contentT->Length());
    contentT->check();
    check();
}

// Copy null-terminated T* array
template <class T>
TupleBase<T>::TupleBase(const T *contentT)
    : _length(T::findNull(contentT))
{
    T::check(contentT, _length);
    T::init(_content, _length);
    T::dup(content(), contentT, _length);
    check();
}

// Repeat T& as array - is there a need for this?
template <class T>
TupleBase<T>::TupleBase(const T &contentT, int repetition)
    : _length(repetition)
{
    T::init(_content, _length);
    T::segFill(content(), contentT, repetition);
    check();
}

// Copy a TupleBase<T>'s contents
template <class T>
TupleBase<T>::TupleBase(const TupleBase<T> *contentT)
    : _length(contentT->Length())
{
    T::check(contentT->content(), contentT->Length());
    T::init(_content, _length);
    T::dup(_content, contentT->content(), _length <? contentT->Length());
    check();
}

template <class T>
TupleBase<T>::TupleBase(const TupleBase<T> &contentT)
    : _length(contentT.Length())
{
    T::check(contentT.content(), contentT.Length());
    T::init(_content, _length);
    T::dup(_content, contentT.content(), _length <? contentT.Length());
    check();
}

// Copy a subrange of TupleBase<T>'s contents
template <class T>
TupleBase<T>::TupleBase(const TupleBase<T> *contentT, int _start, int l)
    : _length(l <? (contentT->Length() - _start))
{
    assert((_length > 0) && (_length <= contentT->Length()));
    T::check(contentT->content(), l);
    T::init(_content, _length);
    T::dup(_content, contentT->content() + _start, l);
    check();
}

template <class T>
TupleBase<T>::TupleBase(const TupleBase<T> &contentT, int _start, int l)
    : _length(l <? (contentT.Length() - _start))
{
    assert((l > 0) && (l <= contentT.Length()));
    T::check(contentT.content(), l);
    T::init(_content, _length);
    T::dup(_content, contentT.content() + _start, l);
    check();
}

// union vector constructor - returns a vector of v1 + v2
template <class T>
TupleBase<T>::TupleBase(const TupleBase<T> *v1, const TupleBase<T> *v2)
    : _length(v1->Length() + v2->Length())
{
    T::check(v1->content(), v1->Length());
    T::check(v2->content(), v2->Length());
    
    T::init(_content, _length);
    
    T::dup(_content, v1->content(), v1->Length());
    T::dup(_content + v1->Length(), v2->content(), v2->Length());
    check();
}

template <class T>
TupleBase<T>::TupleBase(const TupleBase<T> &v1, const TupleBase<T> &v2)
    : _length(v1.Length() + v2.Length())
{
    T::check(v1.content(), v1.Length());
    T::check(v2.content(), v2.Length());
    
    T::init(_content, _length);
    
    T::dup(_content, v1.content(), v1.Length());
    T::dup(_content + v1.Length(), v2.content(), v2.Length());
    check();
}

// replace range with array
template <class T>
TupleBase<T> *TupleBase<T>::Replace(int where, int what, const T *addT, int additional)
{
    assert(what > 0);
    check();
    T::check(addT, what);
    
    if ((where = endRelative(where)) < 0)
        throw new Error("range", where, "start before 0");
    
    // check bounds on insertion point
    if (where >= _length)
        throw new Error("range", where, "start after end");
    
    if ((int)_length - where <= additional)
        throw new Error("range", where, "replacement size");
    
    T::dup(_content + where, addT, what);
    
    check();
    return this;
}

// check Tuplebase's contents
template <class T>
void TupleBase<T>::check(int n) const
{
    int bound = bounds();
#ifdef DEBUGLOG
    cerr << "TupleCheck " << bound
         << " length: " << _length
         << " Memory::mySize: " << Memory::mySize()
         << " offset: " << offsetof(TupleBase<T>, _content)
         << " sizeof: " << sizeof(T)
         << '\n';
#endif
    assert(bound >= _length);
#ifdef BTREE_INTEGRITY
    Memory::Assert();
#endif
    T::check(content(), Length());
}

// dump TupleBase to file
template <class T>
ostream& operator<< (ostream& out,  const TupleBase<T> & vec)
{
    if (&vec) {
        out.form("[0x%08x,%d] ",
                 vec.content(), vec.Length());
        out << "\"";
        const T *contents = vec.content();
        for (int i = 0; i < vec.Length(); i++)
            out << contents[i];
        out << "\"";
    } else {
        out << "[NULL]";
    }
    return out;
}

#endif
