// Q - Simple doubly-linked list class
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms
//	$Id$

#ifndef Q_HH
#define Q_HH
#include <assert.h>

template <class T>
class Qh 
{
    T *start;

public:
    Qh(): start(NULL) {}

    T *operator =(T *element) {
        DEBLOG(cerr << '<' << this << " Qh = " << element << "> ");
        return start = element;
    }

    bool operator ==(T *element) {
        return start == element;
    }

    operator T*() {
        //DEBLOG(cerr << this << " Qh as " << (T*)start);
        return (T*)start;
    }

    // Predicates:
    bool empty() const {
        DEBLOG(cerr << "(Qh " << this << ((start==NULL)?" empty) ":" non-empty) "));
        return start == NULL;
    }
    operator bool() {
        DEBLOG(cerr << "(Qh " << this << ((start==NULL)?" empty) ":" non-empty) "));
        return start != NULL;
    }    
};

/** class Q<T> - doubly linked list class
 *
 * a Q<T> is a triple [Q<T> *prev, Q<T> *next, Q<T> *parent]
 * such that prev & next != NULL
 * an empty Q<T> is a Q<T> in which prev == next == this.
 * Invariant:  next.prev = this, prev.next = this
 *
 * each Q is (optionally) a member of a group of Qs belonging to a parent
 * 
 */ 
template <class T>
class Q
{
protected:
    Q<T> *_next;
    Q<T> *_prev;
    Qh<T> &_parent;	// pointer somewhere into this ring

public:
    // Constructors

    // Default Constructor
    /** construct an empty Q<T>
     */
    Q(Qh<T> &parent)
        : _next(this), _prev(this), _parent(parent)
        {
            DEBLOG(cerr << this << " Q on chain " << &parent << ' ');
            if (!_parent) {
                _parent = (T*)this;
                DEBLOG(cerr << '(' << _prev << ',' <<_next << ") in " << &_parent << '\n');
            } else {
                ((T*)_parent)->Q<T>::add(this);
                DEBLOG(cerr << '(' << _prev << ',' <<_next << ") in " << &_parent << '\n');
            }
        }

    virtual ~Q() { del(); }
        
    // Constructor
    Q(T *q)
        : _next(this), _prev(this), _parent(q->_parent)
        {
            DEBLOG(cerr << this << " Q after " << q << ' ');
            if (q)
                q->Q<T>::add(this);	// insert this at end of q
            if (!_parent)
                _parent = (T*)this;
            DEBLOG(cerr << '(' << _prev << ',' <<_next << ") in " << &_parent << '\n');
        }
    
    // Copy Constructor
    // it doesn't actually copy, as that would be fatal
    Q(T &q)
        : _next(this), _prev(this), _parent(((Q<T>&)q)._parent)
        {
            DEBLOG(cerr << this << " Q copy " << &q << ' ');
            q.Q<T>::add(this);	// insert this at end of q
            if (!_parent)
                _parent = (T*)this;
            DEBLOG(cerr << '(' << _prev << ',' <<_next << ") in " << &_parent << '\n');
        }
    
    // Accessors
    T *next() const {
        return (T*)_next;
    }
    
    T *prev() const {
        return (T*)_prev;
    }
    
    // Predicates:
    bool empty() const {
        return !this || (this == _next);
    }
    
    // Invariant:  next.prev = this, prev.next = this
    void check() const {
        assert ((_prev->_next == this) && (_next->_prev == this));
    }
    
    // Mutators
    
    // delete this from whatever Q<T> it's on
    //	returns an empty list
    void del() {
        check();
        
        if (!empty()) {
            if (_parent == this) {
                _parent = (T*)_next;
            }

            _next->_prev = _prev;
            _prev->_next = _next;
            _next = _prev = this;	// make this an empty list
        } else {
            if (_parent == this) {
                _parent = NULL;
            }
        }
    }
    
    // add q after this
    Q<T> *add(Q<T> *q) {
        check();
        q->check();
        
        q->del();		// delete q from any q it's on
        
        _next->_prev = q;	// our former _next now points back to q
        q->_next =_next;	// q points forward to our former _next
        q->_prev = this;	// q points backward to this
        _next = q;		// q is our current _next
        q->_parent = _parent;	// q's on our ring
        return this;
    }
    
    // insert q before this
    void ins(Q<T> *q) {
        check();
        q->check();

        q->del();		// delete q from any q it's on

        _prev->_next = q;	// our former _prev now points forward to q
        q->_prev = _prev;	// q points back to our former _prev
        q->_next = this;	// q points forward to this
        _prev = q;		// q is our current _prev
        q->_parent = _parent;	// q's on our ring
    }
    
#if 0
    operator T *() {
        return (T*)this;
    }
#endif
};

// Q iterator
template <class T>
class Qi
{
    T *_head;
    T *_curr;

public:
#if 0
    Qi(Q<T> *q)
        : _head(NULL), _curr(NULL)
        {
            if (q) {
                _curr = q->next();
                q->Q<T>::check();
            }
        }
#endif

    Qi(T *q)
        : _head(NULL), _curr(NULL)
        {
            if (q) {
                _curr = q;
                q->Q<T>::check();
            }
        }

    bool check() {
        _curr->check();
        _head->check();
        return true;
    }
    
    Q<T> *Next() {
        if (!_head)
            _head = _curr;
        return (_curr = _curr->Q<T>::next());
    }
    
    Q<T> *Prev() {
        if (!_head)
            _head = _curr;
        return (_curr = _curr->Q<T>::prev());
    }
    
    bool More() const {
        return (_head != _curr);
    }
    
    operator bool() const {
        return (_head != _curr);
    }
    
    Q<T> *operator ++(int) {
        return Next();
    }
    
    Q<T> *operator --(int) {
        return Prev();
    }
    
    Q<T> *del() {
        T *newc = (T*)(_curr->Q<T>::next());
        T *oldc = _curr;
        _curr->Q<T>::del();
        _curr = newc;
        return oldc;
    }
    
    void add(T *q) {
        _curr->Q<T>::add(q);
    }
    
    void ins(T *q) {
        _curr->Q<T>::ins(q);
    }
    
    operator T *() const {
        return _curr;
    }

    T *Current() const {
        return _curr;
    }
};

#endif
