// ElfExe - adding symbols to running Elf processes
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms

static char *id __attribute__((unused))="$Id$";

#include <stdlib.h>
#include <iostream.h>
#include <dlfcn.h>
#include <elf.h>                // AKPM
#include <sys/elf.h>            // AKPM
#include <libelf.h>
#include "elf.hh"
#include "elfexe.hh"

#ifdef DEBUGLOG
#include "elfdump.hh"
#endif

ElfExe::ElfExe()
    : cursymcnt(0), strtablen(0),
      _so((link_map*)dlopen(NULL, RTLD_GLOBAL | RTLD_NOW))
{}

ElfExe::~ElfExe()
{
    // have to reinstate old symtab
    _so->l_info[DT_SYMTAB]->d_un.d_ptr = orgSymTab;
    _so->l_buckets = orgBuckets;
    _so->l_chain = orgChain;
    _so->l_nbuckets = orgNbuckets;
    _so->l_info[DT_HASH]->d_un.d_ptr = orgHash;

    // and old strtab
    _so->l_info[DT_STRTAB]->d_un.d_ptr = (Elf32_Addr)orgStrTab;
}

unsigned char *ElfExe::dupStrTab()
{
    // duplicate strtab
    unsigned char *newstrtab
        = (unsigned char *)malloc(_so->l_info[DT_STRSZ]->d_un.d_val + STRTINC);
    const unsigned char *strtab
        = (unsigned char *)(_so->l_info[DT_STRTAB]->d_un.d_ptr);

#ifdef DEBUGLOG
    cerr << "old strtab: " << (void*)strtab
         << " new strtab: " << (void*)newstrtab
         << '\n';
#endif

    // replace old strtab
    memcpy(newstrtab, strtab, _so->l_info[DT_STRSZ]->d_un.d_val);
    _so->l_info[DT_STRTAB]->d_un.d_ptr = (Elf32_Addr)newstrtab;

    if (strtablen != 0) {
        // the old strtab was malloced - free it
        free((void *)strtab);
    } else {
        // remember the original strtab
        orgStrTab = (ElfW(Addr))strtab;
    }
    strtablen = _so->l_info[DT_STRSZ]->d_un.d_val + STRTINC; // remember len
    return newstrtab;
}

size_t ElfExe::addStrTab(const char *str)
{
    size_t len = strlen((const char *)str);
    size_t end = _so->l_info[DT_STRSZ]->d_un.d_val;
    if (end + len >= strtablen) {
        dupStrTab();	// add some more space
    }
    unsigned char *strtab
        = (unsigned char *)(_so->l_info[DT_STRTAB]->d_un.d_ptr);
    strcpy((char *)strtab + end, (const char *)str);
    _so->l_info[DT_STRSZ]->d_un.d_val += len + 1;
    return end;
}

/* Standard ELF hash function.  Do not change this function; you will
   cause invalid hash tables to be generated.  (Well, you would if this
   were being used yet.)  */
unsigned long ElfExe::elf_hash (const unsigned char *name)
{
    unsigned long h = 0;
    unsigned long g;
    int ch;

    while ((ch = *name++) != '\0') {
        h = (h << 4) + ch;
        if ((g = (h & 0xf0000000)) != 0) {
            h ^= g >> 24;
            h &= ~g;
	}
    }
    return h;
}

/* Array used to determine the number of hash table buckets to use
   based on the number of symbols there are.  If there are fewer than
   3 symbols we use 1 bucket, fewer than 17 symbols we use 3 buckets,
   fewer than 37 we use 17 buckets, and so forth.  We never use more
   than 32771 buckets.  */

const size_t ElfExe::elf_buckets[] =
{
    1, 3, 17, 37, 67, 97, 131, 197, 263, 521, 1031, 2053, 4099, 8209,
    16411, 32771, 0
};

// number of entries in the symtab
unsigned int ElfExe::nrChain()
{
    int *hash = (int *)(_so->l_info[DT_HASH]->d_un.d_ptr + _so->l_addr);
    return hash[1];
}

ElfW(Symndx) *ElfExe::mkShadow(int additional)
{
    size_t dynsymcount;
    int bucketcount = 1;
    bool dynamicST = false;

    // get current occupancy of l
    if (cursymcnt) {
        dynamicST = true;
    } else {
        dynamicST = false;
        cursymcnt = nrChain();
    }
    dynsymcount = cursymcnt + additional;

#ifdef DEBUGLOG
    cerr << "dynsymcount: " << dynsymcount << '\n';
#endif

    // work out size of buckets
    for (int i = 0; elf_buckets[i] != 0; i++) {
        bucketcount = elf_buckets[i];
        if (dynsymcount < elf_buckets[i + 1])
            break;
    }
#ifdef DEBUGLOG
    cerr << "bucketcount: " << bucketcount << '\n';
#endif

    // allocate a new symbol table
    int m_size = ((2 + bucketcount + dynsymcount)
                  * sizeof(ElfW(Symndx)))
        + (dynsymcount * sizeof(ElfW(Sym)));
        ElfW(Symndx) *hash = (ElfW(Symndx) *)malloc(m_size);
    if (!hash)
        throw "malloc null creating shadow hash table";

    memset(hash, 0, m_size);
    hash[0] = bucketcount;
    hash[1] = dynsymcount;	// total space for symbols

#ifdef DEBUGLOG
    cerr << "hash: " << hash << '\n';
#endif

    // add all old symbols to new symbol table.
    const unsigned char *strtab
        = (unsigned char *)(_so->l_info[DT_STRTAB]->d_un.d_ptr);
    ElfW(Symndx) *buckets = hash + 2;
    ElfW(Symndx) *chains = buckets + bucketcount;
    ElfW(Sym) *symtab = (ElfW(Sym)*)(chains + dynsymcount);
    ElfW(Sym) *osymtab = (ElfW(Sym) *)(_so->l_info[DT_SYMTAB]->d_un.d_ptr);
    int chain = 0;

#ifdef DEBUGLOG
    cerr << "buckets: " << buckets << '\n';
    cerr << "chains: " << chains << '\n';
    cerr << "symtab: " << symtab << '\n';
    cerr << "osymtab: " << osymtab << '\n';
#endif
    
    // take each symbol and move it to the new symtab
    symtab[0] = osymtab[0];
    chains[0] = 0;
    for (ElfW(Symndx) symidx = 1; symidx < cursymcnt; symidx++) {
        ElfW(Sym) *sym = osymtab + symidx;
        const unsigned char *name = strtab + sym->st_name;
        
        int hashv = elf_hash(name) % bucketcount;
        ElfW(Symndx) *hp;
        
        for (hp = buckets + hashv; *hp; hp = chains + *hp) {
            if (!*hp)
                break;
        }
        *hp = ++chain;
        symtab[chain] = *sym;
        //cout << "NEWSYM"; dumpSym(l, symtab + chain);        
    }

    hash[0] = bucketcount;
    hash[1] = dynsymcount;	// total space for symbols

    if (dynamicST) {
        // free the old symtab
        free((void*)(_so->l_info[DT_SYMTAB]->d_un.d_ptr));
    } else {
        // remember the original symtab
        orgSymTab = (ElfW(Addr))osymtab;
        orgBuckets = _so->l_buckets;
        orgChain = _so->l_chain;
        orgNbuckets = _so->l_nbuckets;
        orgHash = _so->l_info[DT_HASH]->d_un.d_ptr;
    }

    // replace old symbol table
    _so->l_info[DT_SYMTAB]->d_un.d_ptr = (Elf32_Addr)symtab;
    _so->l_buckets = buckets;
    _so->l_chain = chains;
    _so->l_nbuckets = bucketcount;
    _so->l_info[DT_HASH]->d_un.d_ptr = (int)hash - _so->l_addr;

    cursymcnt = chain;		// remember the occupancy

    return hash;
}

void ElfExe::addSym(const char *name,
            void *value,
            int size,
            int type,
            int bind,
            int section)
{
    // construct the symbol
    ElfW(Sym) sym;
    sym.st_name = addStrTab(name);
    sym.st_size = size;
    sym.st_value = (Elf32_Addr)value;
    sym.st_info = ELF32_ST_INFO(bind, type);
    sym.st_shndx = section;

#ifdef DEBUGLOG
    cerr << "addSym: " << name << '\n';
    dumpSym(_so, &sym);
#endif    

    if (!cursymcnt || (cursymcnt >= nrChain())) {
        mkShadow(SYMTINC);
    }

    // splice the symbol into the hash
    ElfW(Sym) *symtab = (ElfW(Sym) *)(_so->l_info[DT_SYMTAB]->d_un.d_ptr);
    int hashv = elf_hash((const unsigned char *)name) % _so->l_nbuckets;
    ElfW(Symndx) *hp;
        
    for (hp = (ElfW(Symndx) *)(_so->l_buckets) + hashv;
               *hp;
         hp = (ElfW(Symndx) *)(_so->l_chain) + *hp)
        {
            if (!*hp)
                break;
        }
    *hp = ++cursymcnt;
    symtab[cursymcnt] = sym;
#ifdef DEBUGLOG
    cerr << "addSym: " << name << '\n';
    dumpSymIdx(_so, cursymcnt);
#endif    
}

#if 0
const unsigned char *findStrTab(link_map *l, const unsigned char *name)
{
    const unsigned char *strtab
        = (unsigned char *)(_so->l_info[DT_STRTAB]->d_un.d_ptr);
    return strtab + strtablen;
}

int main(int argc, char **argv)
{
    link_map *lexe = (link_map*)dlopen(NULL, RTLD_GLOBAL | RTLD_NOW);
    addSym(lexe, "fred", (void*)0x123456, 4, STT_OBJECT, STB_GLOBAL);
    void *sym = dlsym(lexe, "fred");
    char *err = dlerror();
    if (err)
        cerr << err << '\n';
    else
        cout << "fred: " << sym << '\n';

    // test that the new symtab works
    sym = dlsym(lexe, "_GLOBAL_OFFSET_TABLE_");
    err = dlerror();
    if (err)
        cerr << err << '\n';

    cout << lexe << '\n';
    for (link_map *l = lexe; l; l = l->l_next) {
        dumpObj(l);
    }

    // dump all the syms in lexe
    unsigned int maxsym = nrChain(lexe);
    for (ElfW(Symndx) symidx = 0; symidx < maxsym; symidx++) {
        cout << symidx
             << ' ' << lexe->l_buckets[symidx]
             << ' ' << lexe->l_chain[symidx]
             << '\n';
        dumpSymIdx(lexe, symidx);
    }
}
#endif
